using System;
using System.IO;			// for StreamWriter I/O to a file
using System.Windows.Forms;	// for RichTextBox

namespace MyNamespace
{
	/// <summary>
	///		Multiple precision floating-point Matrix algorithms
	///		in Visual C# .Net
	///
	/// ---------------- *** LEGAL NOTICE: *** ------------------
	///
	/// This program is free software; you can redistribute it and/or modify
	/// it under the terms of the GNU General Public License (GPL) as published
	/// by the Free Software Foundation. cf. the file COPYING.txt.
	///
	/// ------------ *** end of legal notice *** ------------------
	/// </summary>

	// ------- A multiple-precision signed floating matrix
	public class MultiM : MultiF	// array of reals
	{
		public int Rs;		// Number of rows
		public int Cs;		// Number of columns
		public MultiF[,] E;	// elements

		public MultiM()				// initialize a scalar with M = 1
			: base()				// calls MultiF()
		{
			// constructor logic
			this.Rs = 0;
			this.Cs = 0;
			this.E = new MultiF[0, 0];
		}

		public MultiM(int nMaxI)		// initialize a scalar with M = nMaxI
			: base(nMaxI)				// calls MultiF(i)
		{
			// constructor logic
			this.Rs = 0;
			this.Cs = 0;
			this.E = new MultiF[0, 0];
		}

		public MultiM(int nMaxI, int rs, int cs)	// initialize a matrix
			: base(0)								// calls MultiF(0)
		{
			// constructor logic
			if (rs < 1 || cs < 1)
			{
				rs = cs = 0;						// change to a scalar
				this.V = new int[nMaxI+1];
				this.M = nMaxI;
				this.N = 1;
			}
			this.Rs = rs;
			this.Cs = cs;
			this.E = new MultiF[rs, cs];
			for (int r = 0; r < rs; r++)
				for (int c = 0; c < cs; c++)
				{
					this.E[r, c] = new MultiF(nMaxI);
				}
		}

		//		public static bool operator = (MultiM left, MultiM right)	// not allowed to overload =
//		{
//			MultiMD.SetToM(right, left);
//			return true;
//		}

		public static MultiM operator & (MultiM left, MultiM right)	// use &= for = assignment
		{
			MultiMD.SetToM(right, left);
			return left;
		}

		public static MultiM operator & (MultiM left, MultiF right)	// use &= for = assignment
		{
			MultiMD.SetToFM(right, left);
			return left;
		}

		public static MultiM operator & (MultiM left, MultiSI right)	// use &= for = assignment
		{
			MultiMD.SetToSIM(right, left);
			return left;
		}

		public static MultiM operator & (MultiM left, int right)	// use &= for = assignment
		{
			MultiMD.SetTo1M(right, left);
			return left;
		}

		public static MultiM operator & (MultiM left, long right)	// use &= for = assignment
		{
			MultiMD.SetToExM(right, left);
			return left;
		}

		public static MultiM operator & (MultiM left, double right)	// use &= for = assignment
		{
			MultiMD.SetToDbM(right, left);
			return left;
		}

		public static MultiM operator & (MultiM left, decimal right)	// use &= for = assignment
		{
			MultiMD.SetToDeM(right, left);
			return left;
		}

		public static MultiM operator + (MultiM left, MultiM right) // can be used for a += b
		{
			MultiM z = new MultiM(Calc.fMC, left.Rs, left.Cs);
			MultiMD.AddM(left, right, z);
			return z;
		}

		public static MultiM operator + (MultiM left, int right) // can be used for a += b
		{
			MultiM z = new MultiM(Calc.fMC);
			MultiMD.Add1M(left, right, z);
			return z;
		}

		public static MultiM operator + (int left, MultiM right) // can be used for a += b
		{
			MultiM z = new MultiM(Calc.fMC);
			MultiMD.Add1M(right, left, z);
			return z;
		}

		public static MultiM operator - (MultiM left, MultiM right) // can be used for a -= b
		{
			MultiM z = new MultiM(Calc.fMC, left.Rs, left.Cs);
			MultiMD.SubM(left, right, z);
			return z;
		}

		public static MultiM operator - (MultiM left, int right) // can be used for a -= b
		{
			MultiM z = new MultiM(Calc.fMC);
			MultiMD.Add1M(left, -right, z);
			return z;
		}

		public static MultiM operator - (int left, MultiM right) // can be used for a -= b
		{
			MultiM z = new MultiM(Calc.fMC);
			MultiMD.Add1M(right, -left, z);
			MultiMD.ChSM(z);
			return z;
		}

		public static MultiM operator - (MultiM x)
		{
			MultiM z = new MultiM(x.N, x.Rs, x.Cs);
			MultiMD.SetToMS(x, z);
			MultiMD.ChSM(z);
			return z;
		}
		public static MultiM operator * (MultiM left, MultiM right) // can be used for a *= b
		{
			MultiM z = new MultiM(Calc.fMC, left.Rs, right.Cs);
			MultiMD.MulM(left, right, z);
			return z;
		}

		public static MultiM operator * (MultiM left, int right) // can be used for a *= b
		{
			MultiM z = new MultiM(Calc.fMC, left.Rs, left.Cs);
			MultiMD.Mul1M(left, right, z);
			return z;
		}

		public static MultiM operator * (int left, MultiM right)
		{
			MultiM z = new MultiM(Calc.fMC, right.Rs, right.Cs);
			MultiMD.Mul1M(right, left, z);
			return z;
		}

		public static MultiM operator / (MultiM left, MultiM right) // can be used for a /= b
		{
			int maxRs = Math.Max(left.Rs, right.Rs);
			int maxCs = Math.Max(left.Cs, right.Cs);
			MultiM q = new MultiM(Calc.fMC, maxRs, maxCs);
			MultiMD.DivM(left, right, q);
			return q;
		}

		public static MultiM operator / (MultiM left, int right) // can be used for a /= b
		{
			MultiM q = new MultiM(Calc.fMC, left.Rs, left.Cs);
			MultiMD.Div1M(left, right, q);
			return q;
		}

		// % and ^ will default to Singed Integer versions

		public static bool operator == (int d1, MultiM z)
		{
			if (z.Rs != 0)
			{
				MultiMD.ErrIncompatible();
				return false;
			}
			MultiM dF = new MultiM(MultiID.sdInInt);
			MultiMD.SetTo1M(d1, dF);
			return (MultiFD.CompFF(dF, z) == 0);
		}

		public static bool operator != (int d1, MultiM z)
		{
			if (z.Rs != 0)
			{
				MultiMD.ErrIncompatible();
				return true;
			}
			MultiM dF = new MultiM(MultiID.sdInInt);
			MultiMD.SetTo1M(d1, dF);
			return (MultiFD.CompFF(dF, z) != 0);
		}

		public static bool operator == (MultiM z, int d1)
		{
			if (z.Rs != 0)
			{
				MultiMD.ErrIncompatible();
				return false;
			}
			MultiM dF = new MultiM(MultiID.sdInInt);
			MultiMD.SetTo1M(d1, dF);
			return (MultiFD.CompFF(dF, z) == 0);
		}

		public static bool operator != (MultiM z, int d1)
		{
			if (z.Rs != 0)
			{
				MultiMD.ErrIncompatible();
				return true;
			}
			MultiM dF = new MultiM(MultiID.sdInInt);
			MultiMD.SetTo1M(d1, dF);
			return (MultiFD.CompFF(dF, z) != 0);
		}

		public static bool operator == (long ex, MultiM z)
		{
			if (z.Rs != 0)
			{
				MultiMD.ErrIncompatible();
				return false;
			}
			MultiM exF = new MultiM(MultiID.sdInLong);
			MultiMD.SetToExM(ex, exF);
			return (MultiFD.CompFF(exF, z) == 0);
		}

		public static bool operator != (long ex, MultiM z)
		{
			if (z.Rs != 0)
			{
				MultiMD.ErrIncompatible();
				return true;
			}
			MultiM exF = new MultiM(MultiID.sdInLong);
			MultiMD.SetToExM(ex, exF);
			return (MultiFD.CompFF(exF, z) != 0);
		}

		public static bool operator == (MultiM z, long ex)
		{
			if (z.Rs != 0)
			{
				MultiMD.ErrIncompatible();
				return false;
			}
			MultiM exF = new MultiM(MultiID.sdInLong);
			MultiMD.SetToExM(ex, exF);
			return (MultiFD.CompFF(exF, z) == 0);
		}

		public static bool operator != (MultiM z, long ex)
		{
			if (z.Rs != 0)
			{
				MultiMD.ErrIncompatible();
				return true;
			}
			MultiM exF = new MultiM(MultiID.sdInLong);
			MultiMD.SetToExM(ex, exF);
			return (MultiFD.CompFF(exF, z) != 0);
		}

		public static bool operator == (MultiM x, MultiM y)
		{
			return MultiMD.EqualM(x, y);
		}

		public static bool operator != (MultiM x, MultiM y)
		{
			return !MultiMD.EqualM(x, y);
		}

		public static bool operator < (MultiM x, MultiM y)
		{
			if (x.Rs != 0 || y.Rs != 0)
			{
				MultiMD.ErrNotScalar();
				return false;
			}
			return (MultiFD.CompFF(y, x) < 0);
		}

		public static bool operator > (MultiM x, MultiM y)
		{
			if (x.Rs != 0 || y.Rs != 0)
			{
				MultiMD.ErrNotScalar();
				return false;
			}
			return (MultiFD.CompFF(y, x) > 0);
		}

		public static bool operator <= (MultiM x, MultiM y)
		{
			if (x.Rs != 0 || y.Rs != 0)
			{
				MultiMD.ErrNotScalar();
				return false;
			}
			return (MultiFD.CompFF(y, x) <= 0);
		}

		public static bool operator >= (MultiM x, MultiM y)
		{
			if (x.Rs != 0 || y.Rs != 0)
			{
				MultiMD.ErrNotScalar();
				return false;
			}
			return (MultiFD.CompFF(y, x) >= 0);
		}

		public static MultiM operator ++ (MultiM x)
		{
			if (x.Rs != 0)
			{
				MultiMD.ErrNotScalar();
				return MultiMD.mZero;
			}
			MultiFD.RAddF(MultiFD.fOne, x);
			return x;
		}

		public static MultiM operator -- (MultiM x)
		{
			if (x.Rs != 0)
			{
				MultiMD.ErrNotScalar();
				return MultiMD.mZero;
			}
			MultiFD.RAddF(MultiFD.fMOne, x);
			return x;
		}

		public override bool Equals(object o)	// dummy
		{
			return true;
		}

		public override int GetHashCode()		// dummy
		{
			return 0;
		}
	}

	public class MultiMD
	{
		public MultiMD()	// constructor logic
		// initialize Multiple-precision floating matrix package
		// input: MultiID.muDMax, MultiID.muNMax, MultiID.baseM
		{
			mZero = new MultiM(1);			// mZero = 0.0
			mOne = new MultiM(1);
			SetTo1M(1, mOne);				// mOne = 1.0
			mTwo = new MultiM(1);
			SetTo1M(2, mTwo);				// mTwo = 2.0
			vZero = new MultiM(1, 1, 1);	// vZero = {0.0}
		}

		public static bool EqualM(MultiM x, MultiM y)
		{
			if (x.Rs == 0 && y.Rs == 0)
				return (MultiFD.CompFF(y, x) == 0);
			if (x.Rs != y.Rs || x.Cs != y.Cs)
			{
				ErrIncompatible();
				return false;
			}
			bool eq;
			for (int r = 0; r < x.Rs; r++)
				for (int c = 0; c < x.Cs; c++)
				{
					eq = (MultiFD.CompFF(x.E[r, c], y.E[r, c]) == 0);
					if (!eq)
						return false;
				}
			return true;
		} // EqualM

		public static void Matrix(int nMaxI, int rs, int cs, out MultiM z) // init matrix
		{
			z = new MultiM(nMaxI, rs, cs);
		}

		public static void DoneM(ref MultiM z) // free up space
		{
			z = new MultiM();
		}

		public static void ReSizeM(int m, ref MultiM z) // change M and set to zero
		{
			z = new MultiM(m, z.Rs, z.Cs);
		}

		public static void ClearM(ref MultiM z)	// set Value = 0
		{
			z = new MultiM(z.M, z.Rs, z.Cs);
		}

		public static void ChSM(MultiM z) // change sign, z = - z
		{
			MultiFD.ChSF(z);
			for (int r = 0; r < z.Rs; r++)
				for (int c = 0; c < z.Cs; c++)
					MultiFD.ChSF(z.E[r, c]);
		}

		public static void SetToM(MultiM x, MultiM z)	// z = x
		// can reduce precision and round
		{
			int rs = x.Rs;
			int cs = x.Cs;
			if (rs != z.Rs || cs != z.Cs)
				ErrIncompatible();
			else
			{
				MultiFD.SetToF(x, z);
				for (int r = 0; r < rs; r++)
				{
					for (int c = 0; c < cs; c++)
					{
						MultiFD.SetToF(x.E[r, c], z.E[r, c]);
					}
				}
			}
		}

		public static void SetToSIM(MultiSI x, MultiM z) // z = x
		// can reduce precision and round
		{
			if (z.Rs > 0)
			{
				z.E = new MultiF[0, 0];
				z.Rs = z.Cs = 0;
			}
			MultiFD.SetToSIF(x, z);
		}

		public static void SetToMS(MultiM x, MultiM z) // z = x, same size
		// copy everything but .Nm Name
		{
			int rs = x.Rs;
			int cs = x.Cs;
			if (rs != z.Rs || cs != z.Cs)
				ErrIncompatible();
			else
			{
				MultiFD.SetToFS(x, z);
				for (int r = 0; r < rs; r++)
				{
					for (int c = 0; c < cs; c++)
					{
						MultiFD.SetToFS(x.E[r, c], z.E[r, c]);
					}
				}
			}
		}

		public static void SetToFM(MultiF x, MultiM z)	// z = x
			// can reduce precision and round
		{
			if (z.Rs > 0)
			{
				z.E = new MultiF[0, 0];
				z.Rs = z.Cs = 0;
			}
			MultiFD.SetToF(x, z);
		}

		public static void NormM(MultiM z) // normalize
		{
			MultiFD.NormF(z);
			for (int r = 0; r < z.Rs; r++)
			{
				for (int c = 0; c < z.Cs; c++)
				{
					MultiFD.NormF(z.E[r, c]);
				}
			}
		}

		public static void RAddM(MultiM x, MultiM z) // z = z + x
		// sets muErr true if overflow
		{
			AddM(z, x, z);
		}

		public static void RSubM(MultiM x, MultiM z) // z = z - x
		// sets muErr true if overflow
		{
			if (Object.ReferenceEquals(x, z))
				ClearM(ref z);
			else
			{
				ChSM(x);
				RAddM(x, z);
				ChSM(x);
			}		
		}

		public static void AddM(MultiM x, MultiM y, MultiM z) // z = x + y
		// sets muErr true if overflow
		{
			if (x.Rs == 0 && y.Rs == 0)
				MultiFD.AddF(x, y, z);
			else if (x.Rs != y.Rs || x.Cs != y.Cs)
				ErrIncompatible();
			else
				for (int r = 0; r < x.Rs; r++)
					for (int c = 0; c < x.Cs; c++)
						z.E[r, c] &= x.E[r, c] + y.E[r, c];
		}

		public static void SubM(MultiM x, MultiM y, MultiM z) // z = x - y
		// sets muErr true if overflow
		{
			if (Object.ReferenceEquals(x, y))
				ClearM(ref z);
			else if (Object.ReferenceEquals(x, z))
				RSubM(y, z);
			else if (Object.ReferenceEquals(y, z))
			{
				RSubM(x, z);
				ChSM(z);
			}
			else
			{
				ChSM(y);
				AddM(x, y, z);
				ChSM(y);
			}
		}

		//public static void ValueM(string st, out int i, MultiM z)
		//// convert String to z, returns i = # of character used
		//// allows st to be a literal, muErr set true if overflow
		//	{
		//		if (z.Rs > 0)
		//		{
		//			z.E = new MultiF[0, 0];
		//			z.Rs = z.Cs = 0;
		//		}
		//		MultiFD.ValueF(st, out i, z);
		//	}

		public static bool aDigit(string st, string maxD)	// true if ch is a digit <= maxD
		{
			char ch;
			ch = Char.ToUpper(st[0]);
			return (ch >= '0' && ch <= maxD[0] && (ch <= '9' || ch >= 'A'));
		}

		public static void SetTo1M(int d1, MultiM z)	// z = d1
		// allows a literal d1, d1 an integer, d1 not changed
		{
			if (z.Rs > 0)
			{
				z.E = new MultiF[0, 0];
				z.Rs = z.Cs = 0;
			}
			MultiFD.SetTo1F(d1, z);
		}

		public static void SetToExM(Int64 ext, MultiM z)	// z = ext (Int64 = long)
		// allows a literal ext, ext not changed, max Int64 = 9223,37203,68547,75807 = 2^63-1
		{
			if (z.Rs > 0)
			{
				z.E = new MultiF[0, 0];
				z.Rs = z.Cs = 0;
			}
			MultiFD.SetToExF(ext, z);
		}

		public static void SetToDbM(double db, MultiM z) // z = db % baseM^4
		// allows a literal db, db not changed, muErr set true if overflow
		{
			if (z.Rs > 0)
			{
				z.E = new MultiF[0, 0];
				z.Rs = z.Cs = 0;
			}
			MultiFD.SetToDbF(db, z);
		}

		public static void SetToDeM(decimal de, MultiM z) // z = db
		// allows a literal de, de not changed, muErr set true if overflow
		{
			if (z.Rs > 0)
			{
				z.E = new MultiF[0, 0];
				z.Rs = z.Cs = 0;
			}
			MultiFD.SetToDeF(de, z);
		}

		public static void Get1M(out int d1, MultiM z) // d1 = z % baseM
		// z is not changed
		{
			d1 = 0;
			if (z.Rs != 0)
				ErrIncompatible();
			else
				MultiFD.Get1F(out d1, z);
		}

		public static void GetDeM(out decimal ext, MultiM z)	// ext = z
		// z is not changed
		// max integer decimal is 2^96-1 = 7922,81625,14264,33759,35439,50335 29 digits
		{
			ext = 0;
			if (z.Rs != 0)
				ErrIncompatible();
			else
				MultiFD.GetDeF(out ext, z);
		}

		public static void GetDbM(out double db, MultiM z)	// db = IntegerPart(z) % baseM^4
		// z is not changed, 2^53-1 = 9,00719,92547,40991 16 digits
		{
			db = 0;
			if (z.Rs != 0)
				ErrIncompatible();
			else
				MultiFD.GetDbF(out db, z);
		}

		public static void GetDbA(out double db, MultiM z)	// db = IntegerPart(z) + FractionPart(z)
		// z is not changed
		{
			db = 0;
			if (z.Rs != 0)
				ErrIncompatible();
			else
				MultiFD.GetDbA(out db, z);
		}

		public static void GetDbInM(out double db, double low, double high, MultiM z)
		// db = z % baseM^4, in a range, + or -
		{
			db = 0;
			if (z.Rs != 0)
				ErrIncompatible();
			else
				MultiFD.GetDbInF(out db, low, high, z);
		}

		public static void RAdd1M(int d1, MultiM z) // z = z + d1
		// allows literal d1, d1 is not changed, sets muErr true if overflow
		{
			if (z.Rs != 0)
				ErrIncompatible();
			else
				MultiFD.RAdd1F(d1, z);
		}

		public static void RSub1M(int d1, MultiM z) // z = z - d1
		// allows literal d1, d1 is not changed, sets muErr true if overflow
		{
			if (z.Rs != 0)
				ErrIncompatible();
			else
				MultiFD.RSub1F(d1, z);
		}

		public static void Add1M(MultiM x, int d1, MultiM z) // z = x + d1
		// allows literal d1, d1 is not changed, sets muErr true if overflow
		{
			if (x.Rs != 0 || z.Rs != 0)
				ErrIncompatible();
			else
				MultiFD.Add1F(x, d1, z);
		}

		public static void WriteFixedL(RichTextBox outB, StreamWriter outF, MultiM x, ref int tot, int width)
		{
			if (x.Rs == 0)
			{
				MultiFD.WriteFixedL(outB, outF, x, ref tot, width);
			}
			else
			{
				bool wB = MultiID.writeBool;
				MultiID.writeBool = false;
				MultiID.WriteT(outB, outF, "{", ref tot, width);
				int rs = x.Rs;
				int cs = x.Cs;
				for (int r = 0; r < rs; r++)
				{
					for (int c = 0; c < cs; c++)
					{
						MultiID.MuInterrupt();		// allow for abort
						if (MultiID.muAbort) break;	// exit for
						MultiFD.WriteFixedL(outB, outF, x.E[r, c], ref tot, width);
						if (MultiID.diagOn && Calc.listOn)
							MultiID.WriteT(" N=" + x.E[r, c].N + " M=" +
								x.E[r, c].M + " U=" + x.E[r, c].V.GetUpperBound(0), ref tot);
						if (r == rs-1 && c == cs-1)
							MultiID.WriteT(outB, outF, "}", ref tot, width);
						else
						{
							if (c == cs-1)
							{
								MultiID.WriteTLn(outB, outF, ";+", ref tot, width);
								MultiID.WriteTLn(outB, outF, "+", ref tot, width);
							}
							else
								MultiID.WriteTLn(outB, outF, ",+", ref tot, width);
						}
					}
					if (MultiID.muAbort) break;	// exit for
				}
				MultiID.writeBool = wB;
			}
		} // WriteFixedL

		public static void WriteFixed(RichTextBox outB, StreamWriter outF, MultiM z, ref int tot, int width)
		{
			int j;
			int k;
			int did;
			int didZ;
			string st;
			did = 0;
			if (z.S) MultiID.WriteT(outB, outF, "-", ref tot, width);
			if (z.N + z.C > 0)
			{
				st = ((int)(MultiID.baseM + z.V[z.N-1])).ToString();
				j = 1;
				do
					j++;
				while (st[j-1] == '0' && j != MultiID.muDMax + 1);
				did = j - 2 - MultiID.muDMax;
				if (z.N + z.C > 1) did -= MultiID.muDMax;
				if (MultiID.muDpG > 0) did = did % MultiID.muDpG;
				if (did < 0) did = did + MultiID.muDpG;
				didZ = did;
				k = MultiID.muDMax + 1;
				j--;
				Write1M(st, ref j, k, ref did, didZ, outB, outF, ref tot, width);
				if (z.N + z.C > 1)
				{
					if (z.N > 1)
						st = ((int)(MultiID.baseM + z.V[z.N-2])).ToString();
					else
					{
						st = ((int)(MultiID.baseM)).ToString();
					}
					k = MultiID.muDMax + 1;
					j = 1;
					Write1M(st, ref j, k, ref did, didZ, outB, outF, ref tot, width);
				}
			}
			else
			{
				MultiID.WriteT(outB, outF, "0", ref tot, width);
			}
			MultiID.WriteT(outB, outF, ".", ref tot, width);
			didZ = did;
			if (-z.C == 2)
			{
				st = ((int)(MultiID.baseM + z.V[1])).ToString();
				k = MultiID.muDMax + 1;
				j = 1;
				Write1M(st, ref j, k, ref did, didZ, outB, outF, ref tot, width);
			}
			if (-z.C > 0)
			{
				st = ((int)(MultiID.baseM + z.V[0])).ToString();
				j = MultiID.muDMax + 2;
				do
					j--;
				while (st[j-1] == '0');
				k = j;
				j = 1;
				Write1M(st, ref j, k, ref did, didZ, outB, outF, ref tot, width);
			}
			else
			{
				MultiID.WriteT(outB, outF, "0", ref tot, width);
			}
			return;
		} // WriteFixed

		private static void Write1M(string st, ref int j, int k, ref int did, int didZ,
			RichTextBox outB, StreamWriter outF, ref int tot, int width)
		// part of WriteFixed
		{
			while (j < k)
			{
				j++;
				if (did != didZ && MultiID.muDpG > 0)
				{
					if (did % MultiID.muDpG == 0) MultiID.WriteT(outB, outF, ",", ref tot, width);
				}
				MultiID.WriteT(outB, outF, st.Substring(j - 1, 1), ref tot, width);
				did++;
			}
			return;
		}

		public static void WriteM(MultiM z, ref int tot)
		{
			WriteM(MultiID.outS, null, z, ref tot, 0);
		}

		public static void WriteM(RichTextBox outB, StreamWriter outF, MultiM x, ref int tot, int width)
		// output x as a line of text, tot = characters on line so far, x is not modified
		// good for any base = baseO [2, 36], base 10 much faster
		{
			if (x.Rs == 0)
			{
				MultiFD.WriteF(outB, outF, x, ref tot, width);
			}
			else
			{
				bool wB = MultiID.writeBool;
				MultiID.writeBool = false;
				MultiID.WriteT(outB, outF, "{", ref tot, width);
				int rs = x.Rs;
				int cs = x.Cs;
				for (int r = 0; r < rs; r++)
				{
					for (int c = 0; c < cs; c++)
					{
						MultiID.MuInterrupt();		// allow for abort
						if (MultiID.muAbort) break;	// exit for
						MultiFD.WriteF(outB, outF, x.E[r, c], ref tot, width);
						if (MultiID.diagOn && Calc.listOn)
							MultiID.WriteT(" N=" + x.E[r, c].N + " M=" +
								x.E[r, c].M + " U=" + x.E[r, c].V.GetUpperBound(0), ref tot);
						if (r == rs-1 && c == cs-1)
							MultiID.WriteT(outB, outF, "}", ref tot, width);
						else
						{
							if (c == cs-1)
							{
								MultiID.WriteTLn(outB, outF, ";+", ref tot, width);
								MultiID.WriteTLn(outB, outF, "+", ref tot, width);
							}
							else
								MultiID.WriteTLn(outB, outF, ",+", ref tot, width);
						}
					}
					if (MultiID.muAbort) break;	// exit for
				}
				MultiID.writeBool = wB;
			}
		} // WriteM

		public static void WriteMLn(RichTextBox outB, StreamWriter outF, MultiM z, ref int tot)
		// output z as a line of text, tot = characters on line so far
		{
			WriteM(outB, outF, z, ref tot, 0);
			MultiID.Write(outB, outF, "\n");
			tot = 0;
		}

		public static void WriteMLn(MultiM z, ref int tot)
		{
			WriteMLn(MultiID.outS, null, z, ref tot);
		}

		public static void WriteMLn(MultiM z)
		{
			int tot = 0;
			WriteMLn(MultiID.outS, null, z, ref tot);
		}

		public static void RMul1M(int d1, MultiM z) // z = z * d1 (d1 <= maxDigit)
		// Multi-precision one super digit multiply d1 is not changed
		// muErr set true if z overflows
		{
			if (z.Rs == 0)
				MultiFD.RMul1F(d1, z);
			else
				for (int r = 0; r < z.Rs; r++)
					for (int c = 0; c < z.Cs; c++)
						MultiFD.RMul1F(d1, z.E[r, c]);
		}

		public static void Mul1M(MultiM x, int d1, MultiM z) // z = x * d1 (d1 <= maxDigit)
		// Multi-precision one super digit multiply x and d1 are not changed
		// muErr set true if z overflows
		{
			SetToM(x, z);
			RMul1M(d1, z);
		}

		public static MultiM Diagonal(MultiM x) // return diagonal matrix from vector x
		{
			if (x.Rs == 0)
				ErrNotMatrix();
			else if (x.Rs != 1 && x.Cs != 1)
				ErrIncompatible();
			else
			{
				int n = Math.Max(x.Rs, x.Cs);
				MultiM t;
				Matrix(Calc.fMC, n, n, out t);
				for (int r = 0; r < n; r++)
					for (int c = 0; c < n; c++)
						t.E[r, c] &= MultiFD.fZero;
				if (x.Rs == 1)
					for (int i = 0; i < n; i++)
						XPCalLst.SetToNM(x.E[0, i], t.E[i, i]);
				else
					for (int i = 0; i < n; i++)
						XPCalLst.SetToNM(x.E[i, 0], t.E[i, i]);
				return t;
			}
			return mZero;
		} // DiagonalC

		public static void DiagonalM(MultiM x, MultiM z) // z = diagonal matrix from vector x
		{
			z &= Diagonal(x);
		} // DiagonalC

		public static void MulM(MultiM x, MultiM y, MultiM z) // z = x * y
		// x or y or both may have the same address in memory as z
		// x, y are not changed unless they share memory with z
		// sets muErr true if overflow
		{
			if (x.Rs == 0 && y.Rs == 0)
				MultiFD.MulF(x, y, z);
			else if (x.Rs == 0)
			{
				MultiM zM = new MultiM(Calc.fMC, y.Rs, y.Cs);
				for (int r = 0; r < y.Rs; r++)
					for (int c = 0; c < y.Cs; c++)
						MultiFD.SetToF(x * y.E[r, c], zM.E[r, c]);
				SetToM(zM, z);
			}
			else if (y.Rs == 0)
			{
				MultiM zM = new MultiM(Calc.fMC, x.Rs, x.Cs);
				for (int r = 0; r < x.Rs; r++)
					for (int c = 0; c < x.Cs; c++)
						MultiFD.SetToF(y * x.E[r, c], zM.E[r, c]);
				SetToM(zM, z);
			}
			else if (x.Cs != y.Rs)
				ErrIncompatible();
			else
			{
				bool saveExp = MultiFD.expanded;
				int saveFMC = Calc.fMC;
				int saveMMax = MultiID.mMax;
				XPCalLst.Expand();
				MultiF pF = new MultiF(Calc.fMC);
				MultiM zM = new MultiM(Calc.fMC, x.Rs, y.Cs);
				for (int r = 0; r < x.Rs; r++)
					for (int c = 0; c < y.Cs; c++)
					{
						pF &= 0;
						for (int i = 0; i < x.Cs; i++)
						{
							pF += x.E[r, i] * y.E[i, c];
						}
						zM.E[r, c] &= pF;
					}
				SetToM(zM, z);
				MultiFD.expanded = saveExp;
				Calc.fMC = saveFMC;
				MultiID.mMax = saveMMax;
			}
		} // MulN

		public static void RMulM(MultiM x, MultiM z) // z = z * x
		// x may have the same address in memory as z
		// x is not changed unless it shares memory with z
		// sets muErr true if overflow
		{
			MulM(z, x, z);
		}

		public static void SqM(MultiM x, MultiM z) // z = x * x
		// x may have the same address in memory as z
		// x is not changed unless it shares memory with z
		// sets muErr true if overflow
		{
			MulM(x, x, z);
		}

		public static void RSqM(MultiM z) // z = z * z
		// sets muErr true if overflow
		{
			MulM(z, z, z);
		}

		public static void RDivMF(MultiF x, MultiM z) // z = z / x
		// sets muErr true if overflow
		{
			DivMF(z, x, z);
		}

		public static void RDiv1M(int d1, MultiM z) // z = z / d1 (d1 <= maxDigit)
		// Multi-precision one super digit divide, d1 is not changed
		// allows literal d1, muErr set true if d1 = 0
		{
			MultiF x = new MultiF(2); // temp
			MultiFD.SetTo1F(d1, x);
			RDivMF(x, z);
		}

		public static void Div1M(MultiM u, int d1, MultiM z) // z = u / d1 (d1 <= maxDigit)
		// Multi-precision one super digit divide, d1 and u not changed
		// u and z may have same address in memory, then u is changed
		// allows literal d1, muErr set true if d1 = 0
		{
			SetToM(u, z);
			RDiv1M(d1, z);
		}

		public static void DivM(MultiM u, MultiM d, MultiM z) // z = u / d
		// u or d or both may have the same address in memory as z
		// u, d are not changed unless they share memory with z
		// sets muErr true if d = 0
		{
			if (u.Rs == 0 && d.Rs == 0)
				MultiFD.DivF(u, d, z);
			else if (d.Rs != d.Cs)
			{
				ErrNotSquare();
			}
			else if (u.Rs > 0 && d.Rs > 0 && (u.Rs != u.Cs || u.Rs != d.Rs))
			{
				ErrIncompatible();
			}
			else if (d.Rs == 0)
			{
				MultiM dInv = new MultiM(Calc.fMC);
				dInv &= MultiFD.fOne / (MultiF)d;
				MulM(u, dInv, z);
			}
			else
			{
				bool saveExp = MultiFD.expanded;
				int saveFMC = Calc.fMC;
				int saveMMax = MultiID.mMax;
				XPCalLst.Expand();
				MultiM dInv = new MultiM(Calc.fMC, d.Rs, d.Cs);
				InvM(d, dInv);
				MulM(u, dInv, z);
				MultiFD.expanded = saveExp;
				Calc.fMC = saveFMC;
				MultiID.mMax = saveMMax;
				NormM(z);
			}
		} // DivM

		public static void InvM(MultiM x, MultiM z) // z = 1 / x
		{
			if (x.Rs == 0)
				MultiFD.DivF(MultiFD.fOne, x, z);
			else if (x.Cs != x.Rs)
				ErrNotSquare();
			else
			{
				bool saveExp = MultiFD.expanded;
				int saveFMC = Calc.fMC;
				int saveMMax = MultiID.mMax;
				XPCalLst.Expand();
				int i, j;
				int n = x.Rs;
				MultiM a = new MultiM(Calc.fMC, n, n);
				a &= x;
				int[] index;
				index = new int[n];
				MultiF d = new MultiF();
				bool sing;
				LUdecomp(a, index, d, out sing); // decompose the matrix just once
				if (sing)
					ErrSingular();
				else
				{
					MultiM y = new MultiM(Calc.fMC, n, n);
					MultiM b = new MultiM(Calc.fMC, n, 1);
					for (j = 0; j < n; j++)				// solve for column j
					{
						for (i = 0; i < n; i++)			// set up identity matrix column j
						{
							if (i == j)
								b.E[i, 0] &= MultiFD.fOne;
							else
								b.E[i, 0] &= MultiFD.fZero;
						}
						LUbackSub(a, index, b);
						for (i = 0; i < n; i++)			// store column j
							y.E[i, j] &= b.E[i, 0];
					}
					SetToM(y, z);
				}
				MultiFD.expanded = saveExp;
				Calc.fMC = saveFMC;
				MultiID.mMax = saveMMax;
				NormM(z);
			}
		} // InvM
		
		public static void LUdecomp(MultiM a, int[] index, MultiF d, out bool sing)
// LU decomposition, sing is singular matrix flag
// call XPCalLst.Expand(); before calling LUdecomp
// this code is adapted from Numerical Recipes in C
		{
			sing = false;
			int n = a.Rs;
			int nm = n - 1;
			int i, j, k;
			int iMax = 0;
			MultiF big = new MultiF(Calc.fMC);
			MultiF dum = new MultiF(Calc.fMC);
			MultiF sum = new MultiF(Calc.fMC);
			MultiF temp = new MultiF(Calc.fMC);
			MultiF[] vv;
			vv = new MultiF[n];
			for (i = 0; i < n; i++)
				vv[i] = new MultiF(Calc.fMC);
			d &= 1;						// no rows interchanged yet
			for (i = 0; i < n; i++)		// loop over rows to get implicit scaling information
			{
				big &= 0;
				for (j = 0; j < n; j++)
				{
					temp &= a.E[i, j];
					temp.S = false;
					if (temp > big) 
						big &= temp;
				}
				if (big == MultiFD.fZero)
				{
					d &= 0;				// singular matrix
					sing = true;
					return;
				}
				vv[i] &= MultiFD.fOne / big;	// save the scaling
			}
			for (j = 0; j < n; j++)				// this is the loop over columns of C
			{
				for (i = 0; i < j; i++)
				{
					sum &= a.E[i, j];
					for (k = 0; k < i; k++)
						sum -= a.E[i, k] * a.E[k, j];
					a.E[i, j] &= sum;
				}
				big &= MultiFD.fZero;			// initialize for the search for largest pivot element
				for (i = j; i < n; i++)
				{
					sum &= a.E[i, j];
					for (k = 0; k < j; k++)
						sum -= a.E[i, k] * a.E[k, j];
					a.E[i, j] &= sum;
					dum &= sum * vv[i];			// figure of merit for the pivot
					dum.S = false;
					if (dum >= big)				// is it better than the best so far?
					{
						big &= dum;
						iMax = i;
					}
				}
				if (j != iMax)					// do we need to interchange rows?
				{
					for (k = 0; k < n; k++)		// yes we do
					{
						dum &= a.E[iMax, k];
						a.E[iMax, k] &= a.E[j, k];
						a.E[j, k] &= dum;
					}
					d.S = !d.S;					// and change the parity of d
					vv[iMax] &= vv[j];			// also interchange the scale factor
				}
				index[j] = iMax;
				if (a.E[j, j] == MultiFD.fZero)
				{
					d &= 0;				// singular matrix
					sing = true;
					return;
				}
				if (j != nm)					// now finally divide by the pivot element
				{
					dum &= MultiFD.fOne / a.E[j, j];	// if the pivot element is zero the matrix is singular
					for (i = j + 1; i < n; i++)			// for some applications on singular matrices it is
						a.E[i, j] *= dum;				// desirable to substitute tiny for zero
				}
			}
			long next;
			long max = Int64.MinValue;
			for (i = 0; i < n; i++)
			{
				if ((MultiF)a.E[i, i] != MultiFD.fZero)
				{
					next = a.E[i, i].N + a.E[i, i].C;
					if (next > max)
						max = next;
				}
			}
			for (i = 0; i < n; i++)
			{
				next = a.E[i, i].N + a.E[i, i].C;
				if (max - next > Calc.fMC - MultiFD.fGuard)
				{
					MultiFD.ClearF(a.E[i, i]);
					sing = true;
				}
			}
		} // LUdecomp

		public static void LUbackSub(MultiM a, int[] index, MultiM b)
			// this code is adapted from Numerical Recipes in C
		{
			int n = a.Rs;
			int nm = n - 1;
			int i, ip, j;
			int ii = -1;
			MultiF sum = new MultiF(Calc.fMC);
			for (i = 0; i < n; i++) 
			{
				ip = index[i];
				sum &= b.E[ip, 0];
				b.E[ip, 0] &= b.E[i, 0];
				if (ii != -1)
					for (j = ii; j < i; j++)
						sum -= a.E[i, j] * b.E[j, 0];
				else if (sum != MultiFD.fZero)
					ii = i;
				b.E[i, 0] &= sum;
			}
			for (i = nm; i >= 0; i--) 
			{
				sum &= b.E[i, 0];
				for (j = i + 1; j < n; j++)
					sum -= a.E[i, j] * b.E[j, 0];
				b.E[i, 0] &= sum / a.E[i, i];
			}
		} // LUbackSub

		public static void DivMF(MultiM u, MultiF d, MultiM z) // z = u / d
		// u or d or both may have the same address in memory as z
		// u, d are not changed unless they share memory with z
		// sets muErr true if d = 0
		{
			bool saveExp = MultiFD.expanded;
			int saveFMC = Calc.fMC;
			int saveMMax = MultiID.mMax;
			XPCalLst.Expand();
			z &= (MultiF)u / d;
			MultiFD.expanded = saveExp;
			Calc.fMC = saveFMC;
			MultiID.mMax = saveMMax;
			NormM(z);
		} // DivMF

		public static void RDivM(MultiM d, MultiM z) // z = z / d
		// d may have the same address in memory as z
		// d is not changed unless it shares memory with z
		// sets muErr true if d = 0
		{
			DivM(z, d, z);
		}

		public static void SqRootM(MultiM x, MultiM z) // z = SqRt(x)
		// x is not changed, it is ok for x and this to have the same location in memory
		// but then x will be changed. Sets muErr True if error
		{	
			MultiFD.SqRootF(x, z);
		}

		public static void RSqRtM(MultiM z) // z = SqRt(z)
		// sets muErr true if error
		{
			SqRootM(z, z);
		}

		public static void IntM(MultiM x, MultiM z) // z = Integer part of x
		{
			MultiFD.IntF(x, z);
		}

		public static void RIntM(MultiM z) // z = Integer part of z
		{
			IntM(z, z);
		}

		public static void FracM(MultiM x, MultiM z) // z = Fractional part of x
		{
			MultiFD.FracF(x, z);
		}

		public static void RFracM(MultiM z) // z = Fractional part of z
		{
			FracM(z, z);
		}

		public static void FloorM(MultiM x, MultiM z) // z = integer <= and nearest to x
		{
			MultiFD.FloorF(x, z);
		}

		public static void LopM(MultiM x, MultiM z) // z = x with one less super digit
		{
			if (x.Rs == 0)
				MultiFD.LopF(x, z);
			else
			{
				long smallC = Int64.MaxValue;
				for (int r = 0; r < x.Rs; r++)
					for (int c = 0; c < x.Cs; c++)
						if (x.E[r, c].C < smallC && x.E[r, c] != MultiFD.fZero)
							smallC = x.E[r, c].C;
				for (int r = 0; r < x.Rs; r++)
					for (int c = 0; c < x.Cs; c++)
						if (x.E[r, c].C == smallC)
							MultiFD.LopF(x.E[r, c], z.E[r, c]);
						else
							MultiFD.SetToF(x.E[r, c], z.E[r, c]);
			}
		} // LopC

		public static void RLopM(MultiM z)	// z = z with one less least significant "digit"
		{
			LopM(z, z);
		}

		public static bool ZTestM(MultiM z) // true if z is Zero
		{
			if (z.Rs == 0)
				return (MultiFD.ZTestF(z));
			for (int r = 0; r < z.Rs; r++)
				for (int c = 0; c < z.Cs; c++)				
				{
					if (!MultiFD.ZTestF(z.E[r, c]))
						return false;
				}
			return true;
		}

		public static void ReducePrecisionM(int ps, MultiM d, MultiM dL)
		// remove lower ps super digits, d and dL may be the same object
		{
			MultiFD.ReducePrecisionF(ps, d, dL);
		}
	
		public static void MulQ(MultiM x, MultiM y, MultiM z) // z = x * y, all quaternions
		{
			if (x.Rs != 4 || x.Cs != 1 || y.Rs != 4 || y.Cs != 1)
				ErrNotQuat();
			else
			{
				bool saveExp = MultiFD.expanded;
				int saveFMC = Calc.fMC;
				int saveMMax = MultiID.mMax;
				XPCalLst.Expand();
				MultiM zM = new MultiM(Calc.fMC, 4, 1);
				zM.E[0, 0] &= x.E[0, 0] * y.E[3, 0] +
					x.E[1, 0] * y.E[2, 0] -
					x.E[2, 0] * y.E[1, 0] +
					x.E[3, 0] * y.E[0, 0];
				zM.E[1, 0] &= -x.E[0, 0] * y.E[2, 0] +
					x.E[1, 0] * y.E[3, 0] +
					x.E[2, 0] * y.E[0, 0] +
					x.E[3, 0] * y.E[1, 0];
				zM.E[2, 0] &= x.E[0, 0] * y.E[1, 0] -
					x.E[1, 0] * y.E[0, 0] +
					x.E[2, 0] * y.E[3, 0] +
					x.E[3, 0] * y.E[2, 0];
				zM.E[3, 0] &= -x.E[0, 0] * y.E[0, 0] -
					x.E[1, 0] * y.E[1, 0] -
					x.E[2, 0] * y.E[2, 0] +
					x.E[3, 0] * y.E[3, 0];
				SetToM(zM, z);
				MultiFD.expanded = saveExp;
				Calc.fMC = saveFMC;
				MultiID.mMax = saveMMax;
			}
		} // MulQ
	
		public static void AugQ(MultiM x, MultiM z) // z = quaternion augment of vector x
		{
			if (x.Rs != 3 || x.Cs != 1 || z.Rs != 4 || z.Cs != 1)
				ErrIncompatible();
			else
			{
				z.E[0, 0] &= x.E[0, 0];
				z.E[1, 0] &= x.E[1, 0];
				z.E[2, 0] &= x.E[2, 0];
				z.E[3, 0] &= MultiFD.fZero;
			}
		} // AugQ
	
		public static void VPartQ(MultiM x, MultiM z) // z = vector part of quaternion x
		{
			if (x.Rs != 4 || x.Cs != 1 || z.Rs != 3 || z.Cs != 1)
				ErrIncompatible();
			else
			{
				z.E[0, 0] &= x.E[0, 0];
				z.E[1, 0] &= x.E[1, 0];
				z.E[2, 0] &= x.E[2, 0];
			}
		} // VPartQ
	
		public static void PrinQ(MultiM x, MultiM z)	// z = principal equivalent quaternion of x
		{
			if (x.Rs != 4 || x.Cs != 1 || z.Rs != 4 || z.Cs != 1)
				ErrNotQuat();
			else
			{
				int nZ = 3;
				while (nZ > 0 && x.E[nZ, 0] == MultiFD.fZero)
					nZ--;
				if (x.E[nZ, 0].S)
					EqQ(x, z);
				else
					z &= x;
			}
		} // PrinQ
	
		public static void MToQ(MultiM x, MultiM z)	// z = quaternion equivalent of matrix x
		{
			if (x.Rs != 3 || x.Cs != 3 || z.Rs != 4 || z.Cs != 1)
				ErrIncompatible();
			else
			{
				bool saveExp = MultiFD.expanded;
				int saveFMC = Calc.fMC;
				int saveMMax = MultiID.mMax;
				XPCalLst.Expand();
				MultiF t0 = new MultiF(Calc.fMC);
				MultiF t1 = new MultiF(Calc.fMC);
				MultiF t2 = new MultiF(Calc.fMC);
				MultiF t3 = new MultiF(Calc.fMC);
				t0 &= MultiFD.fOne + x.E[0, 0] - x.E[1, 1] - x.E[2, 2];
				t1 &= MultiFD.fOne - x.E[0, 0] + x.E[1, 1] - x.E[2, 2];
				t2 &= MultiFD.fOne - x.E[0, 0] - x.E[1, 1] + x.E[2, 2];
				t3 &= MultiFD.fOne + x.E[0, 0] + x.E[1, 1] + x.E[2, 2];
				if (t0.S || t1.S || t2.S || t3.S)
					ErrNotRotationM();
				else
				{
					int max = 0;
					if (t0 < t1) { max = 1; t0 &= t1; }
					if (t0 < t2) { max = 2; t0 &= t2; }
					if (t0 < t3) { max = 3; t0 &= t3; }
					MultiFD.SqRootF(t0, t0);
					t1 &= t0 / MultiFD.fTwo;
					t2 &= t0 * MultiFD.fTwo;
					if (max == 0)
					{
						z.E[0, 0] &= t1;
						z.E[1, 0] &= (x.E[0, 1] + x.E[1, 0]) / t2;
						z.E[2, 0] &= (x.E[2, 0] + x.E[0, 2]) / t2;
						z.E[3, 0] &= (x.E[1, 2] - x.E[2, 1]) / t2;
					}
					if (max == 1)
					{
						z.E[0, 0] &= (x.E[0, 1] + x.E[1, 0]) / t2;
						z.E[1, 0] &= t1;
						z.E[2, 0] &= (x.E[1, 2] + x.E[2, 1]) / t2;
						z.E[3, 0] &= (x.E[2, 0] - x.E[0, 2]) / t2;
					}
					if (max == 2)
					{
						z.E[0, 0] &= (x.E[2, 0] + x.E[0, 2]) / t2;
						z.E[1, 0] &= (x.E[1, 2] + x.E[2, 1]) / t2;
						z.E[2, 0] &= t1;
						z.E[3, 0] &= (x.E[0, 1] - x.E[1, 0]) / t2;
					}
					if (max == 3)
					{
						z.E[0, 0] &= (x.E[1, 2] - x.E[2, 1]) / t2;
						z.E[1, 0] &= (x.E[2, 0] - x.E[0, 2]) / t2;
						z.E[2, 0] &= (x.E[0, 1] - x.E[1, 0]) / t2;
						z.E[3, 0] &= t1;
					}
					PrinQ(z, z);
				}
				MultiFD.expanded = saveExp;
				Calc.fMC = saveFMC;
				MultiID.mMax = saveMMax;
			}
		} // MToQ
	
		public static void QToM(MultiM x, MultiM z)	// z = matrix equivalent of quaternion x
		{
			if (x.Rs != 4 || x.Cs != 1 || z.Rs != 3 || z.Cs != 3)
				ErrIncompatible();
			else
			{
				bool saveExp = MultiFD.expanded;
				int saveFMC = Calc.fMC;
				int saveMMax = MultiID.mMax;
				XPCalLst.Expand();
				z.E[0, 0] &= x.E[0, 0] * x.E[0, 0] - x.E[1, 0] * x.E[1, 0] -
					x.E[2, 0] * x.E[2, 0] + x.E[3, 0] * x.E[3, 0];
				z.E[1, 1] &= -x.E[0, 0] * x.E[0, 0] + x.E[1, 0] * x.E[1, 0] -
					x.E[2, 0] * x.E[2, 0] + x.E[3, 0] * x.E[3, 0];
				z.E[2, 2] &= -x.E[0, 0] * x.E[0, 0] - x.E[1, 0] * x.E[1, 0] +
					x.E[2, 0] * x.E[2, 0] + x.E[3, 0] * x.E[3, 0];
				z.E[0, 1] &= (x.E[0, 0] * x.E[1, 0] + x.E[2, 0] * x.E[3, 0]) * MultiFD.fTwo;
				z.E[0, 2] &= (x.E[0, 0] * x.E[2, 0] - x.E[1, 0] * x.E[3, 0]) * MultiFD.fTwo;
				z.E[1, 0] &= (x.E[0, 0] * x.E[1, 0] - x.E[2, 0] * x.E[3, 0]) * MultiFD.fTwo;
				z.E[1, 2] &= (x.E[1, 0] * x.E[2, 0] + x.E[0, 0] * x.E[3, 0]) * MultiFD.fTwo;
				z.E[2, 0] &= (x.E[0, 0] * x.E[2, 0] + x.E[1, 0] * x.E[3, 0]) * MultiFD.fTwo;
				z.E[2, 1] &= (x.E[1, 0] * x.E[2, 0] - x.E[0, 0] * x.E[3, 0]) * MultiFD.fTwo;
				MultiFD.expanded = saveExp;
				Calc.fMC = saveFMC;
				MultiID.mMax = saveMMax;
			}
		} // QToM
	
		public static void QToR(MultiM x, MultiM z)	// // z = rotation vector from rotation quaternion
		{
			if (x.Rs != 4 || x.Cs != 1 || z.Rs != 4 || z.Cs != 1)
				ErrIncompatible();
			else
			{
				bool saveExp = MultiFD.expanded;
				int saveFMC = Calc.fMC;
				int saveMMax = MultiID.mMax;
				XPCalLst.Expand();
				MultiF n3 = new MultiF(Calc.fMC);
				MultiM a = new MultiM(Calc.fMC);
				n3 &= x.E[0, 0] * x.E[0, 0] + x.E[1, 0] * x.E[1, 0] + x.E[2, 0] * x.E[2, 0];
				MultiFD.RSqRtF(n3);
				if (n3 > MultiFD.fOne)
					n3 &= MultiFD.fOne;
				int tI;
				Common.InitNAdd(1, "QToR-tI", out tI);
				XPCalLst.ASinN(n3, tI);
				if (x.E[3, 0].S)
				{
					if (Calc.degMode)
					{
						XPCalLst.AddN(180 * MultiFD.fOne, -Common.nList[tI], tI);
					}
					else
					{
						int pI;
						pI = XPCalLst.GetPiPtr();
						XPCalLst.AddN(Common.nList[pI], -Common.nList[tI], tI);
					}
				}
				if (n3 == MultiFD.fZero)
				{
					z.E[0, 0] &= MultiFD.fZero;
					z.E[1, 0] &= MultiFD.fZero;
					z.E[2, 0] &= MultiFD.fZero;
				}
				else
				{
					z.E[0, 0] &= x.E[0, 0] / n3;
					z.E[1, 0] &= x.E[1, 0] / n3;
					z.E[2, 0] &= x.E[2, 0] / n3;
				}
				z.E[3, 0] &= Common.nList[tI] * MultiFD.fTwo;
				Common.LDelete(tI);
				MultiFD.expanded = saveExp;
				Calc.fMC = saveFMC;
				MultiID.mMax = saveMMax;
			}
		} // QToR
	
		public static void RToQ(MultiM x, MultiM z)	// z = rotation quaternion from rotation vector
		{
			if (x.Rs != 4 || x.Cs != 1 || z.Rs != 4 || z.Cs != 1)
				ErrIncompatible();
			else
			{
				bool saveExp = MultiFD.expanded;
				int saveFMC = Calc.fMC;
				int saveMMax = MultiID.mMax;
				XPCalLst.Expand();
				MultiF t = new MultiF(Calc.fMC);
				int tI;
				Common.InitNAdd(1, "RToQ-tI", out tI);
				t &= x.E[3, 0] / MultiFD.fTwo;
				XPCalLst.SinN(t, tI);
				z.E[0, 0] &= Common.nList[tI] * x.E[0, 0];
				z.E[1, 0] &= Common.nList[tI] * x.E[1, 0];
				z.E[2, 0] &= Common.nList[tI] * x.E[2, 0];
				XPCalLst.CosN(t, tI);
				z.E[3, 0] &= Common.nList[tI];
				MultiFD.expanded = saveExp;
				Calc.fMC = saveFMC;
				MultiID.mMax = saveMMax;
			}
		} // RToQ
	
		public static void MToR(MultiM x, MultiM z)	// // z = rotation vector from rotation matrix
		{
			if (x.Rs != 3 || x.Cs != 3 || z.Rs != 4 || z.Cs != 1)
				ErrIncompatible();
			else
			{
				MToQ(x, z);
				QToR(z, z);
			}
		} // MToR
	
		public static void RToM(MultiM x, MultiM z)	// z = rotation matrix from rotation vector
		{
			if (x.Rs != 4 || x.Cs != 1 || z.Rs != 3 || z.Cs != 3)
				ErrIncompatible();
			else
			{
				bool saveExp = MultiFD.expanded;
				int saveFMC = Calc.fMC;
				int saveMMax = MultiID.mMax;
				XPCalLst.Expand();
				MultiM q = new MultiM(Calc.fMC, 4, 1);
				RToQ(x, q);
				QToM(q, z);
				MultiFD.expanded = saveExp;
				Calc.fMC = saveFMC;
				MultiID.mMax = saveMMax;
			}
		} // RToM
	
		public static void RotateQ(MultiM x, MultiM y, MultiM z)
		// z = 3-D vector y rotated by quaternion x
		{
			if (x.Rs != 4 || x.Cs != 1 || y.Rs != 3 || y.Cs != 1 || z.Rs != 3 || z.Cs != 1)
				ErrNotQuat();
			else
			{
				MultiM yA = new MultiM(Calc.fMC, 4, 1);
				MultiM xC = new MultiM(Calc.fMC, 4, 1);
				AugQ(y, yA);
				InvQ(x, xC);
				MulQ(xC, yA, xC);
				MulQ(xC, x, xC);
				VPartQ(xC, z);
			}
		} // RotateQ
	
		public static void RotateM(MultiM x, MultiM y, MultiM z)
		// z = 3-D vector y rotated by matrix x
		{
			if (x.Rs != 3 || x.Cs != 3 || y.Rs != 3 || y.Cs != 1 || z.Rs != 3 || z.Cs != 1)
				ErrIncompatible();
			else
				MulM(x, y, z); 
		} // RotateM
	
		public static void EqQ(MultiM x, MultiM z) // z = quaternion equivalent to x
		{
			if (x.Rs != 4 || x.Cs != 1 || z.Rs != 4 || z.Cs != 1)
				ErrNotQuat();
			else
				z &= -x;
		} // EqQ
	
		public static void InvQ(MultiM x, MultiM z) // z = inverse of quaternion x
		{
			EqQ(x, z);
			MultiFD.ChSF(z.E[3, 0]);
		} // InvQ
	
		public static void NormQ(MultiM x, MultiM z) // z = quaternion x normalized
		{
			if (x.Rs != 4 || x.Cs != 1 || z.Rs != 4 || z.Cs != 1)
				ErrNotQuat();
			else
			{
				bool saveExp = MultiFD.expanded;
				int saveFMC = Calc.fMC;
				int saveMMax = MultiID.mMax;
				XPCalLst.Expand();
				MultiM zM = new MultiM(Calc.fMC, 4, 1);
				MultiM nX = new MultiM(Calc.fMC);
				NormV(x, nX);
				if (nX != mZero)
					zM &= x / nX;
				SetToM(zM, z);
				MultiFD.expanded = saveExp;
				Calc.fMC = saveFMC;
				MultiID.mMax = saveMMax;
			}
		} // NormQ
	
		public static void NormR(MultiM x, MultiM z) // z = rotation vector x normalized
		{
			if (x.Rs != 4 || x.Cs != 1 || z.Rs != 4 || z.Cs != 1)
				ErrNotQuat();
			else
			{
				bool saveExp = MultiFD.expanded;
				int saveFMC = Calc.fMC;
				int saveMMax = MultiID.mMax;
				XPCalLst.Expand();
				MultiM zM = new MultiM(Calc.fMC, 4, 1);
				MultiM nX = new MultiM(Calc.fMC);
				zM &= x;
				zM.E[3, 0] &= mZero;
				NormV(zM, nX);
				if (nX != mZero)
				{
					zM &= zM / nX;
					if (Calc.degMode)
						XPCalLst.Mod(x.E[3, 0], 180 * MultiFD.fTwo, zM.E[3, 0]);
					else
					{
						int pI;
						pI = XPCalLst.GetPiPtr();
						XPCalLst.Mod(x.E[3, 0], Common.nList[pI] * MultiFD.fTwo, zM.E[3, 0]);
					}
				}
				SetToM(zM, z);
				MultiFD.expanded = saveExp;
				Calc.fMC = saveFMC;
				MultiID.mMax = saveMMax;
			}
		} // NormR
	
		public static void NormV(MultiM x, MultiM z) // z = norm of vector x
		{
			if (x.Rs != 1 && x.Cs != 1)
				ErrNotVector();
			else if (z.Rs != 0)
				ErrNotScalar();
			else
			{
				bool saveExp = MultiFD.expanded;
				int saveFMC = Calc.fMC;
				int saveMMax = MultiID.mMax;
				XPCalLst.Expand();
				MultiF zF = new MultiF(Calc.fMC);
				if (x.Rs == 1)
					for (int i = 0; i < x.Cs; i++)
						zF+= x.E[0, i] * x.E[0, i];
				else
					for (int i = 0; i < x.Rs; i++)
						zF += x.E[i, 0] * x.E[i, 0];
				MultiFD.RSqRtF(zF);
				SetToFM(zF, z);
				MultiFD.expanded = saveExp;
				Calc.fMC = saveFMC;
				MultiID.mMax = saveMMax;
			}
		} // NormV
	
		public static void ErrElementOfScalar()
		{
			MultiID.MuWriteErr("Cannot set an element of a scalar");
		}
	
		public static void ErrExponent()
		{
			MultiID.MuWriteErr("Cannot use a matrix as an exponent");
		}
	
		public static void ErrNotScalar()
		{
			MultiID.MuWriteErr("Cannot use a matrix as a scalar");
		}
	
		public static void ErrNotMatrix()
		{
			MultiID.MuWriteErr("Cannot use a scalar as a matrix");
		}
	
		public static void ErrOutside()
		{
			MultiID.MuWriteErr("Element outside of matrix");
		}
	
		public static void ErrIncompatible()
		{
			MultiID.MuWriteErr("Matrices are incompatible");
		}
	
		public static void ErrSingular()
		{
			MultiID.MuWriteErr("Matrix is singular");
		}
	
		public static void ErrNotSquare()
		{
			MultiID.MuWriteErr("This is not a square matrix");
		}
	
		public static void ErrNotVector()
		{
			MultiID.MuWriteErr("This is not a vector");
		}
	
		public static void ErrNotColumn()
		{
			MultiID.MuWriteErr("This is not a column vector");
		}
	
		public static void ErrNotQuat()
		{
			MultiID.MuWriteErr("This is not a quaternion");
		}
	
		public static void ErrNotRotationM()
		{
			MultiID.MuWriteErr("This is not a rotation matrix");
		}
	
		public static MultiM mZero;		// mZero = 0.0 zero scalar
		public static MultiM mOne;		// mOne  = 1.0
		public static MultiM mTwo;		// mTwo  = 2.0
		public static MultiM vZero;		// vZero = {0.0} zero vector
	}
}
